import click


def log_modify_file(file_path):
    click.secho(f"  📁 Modifying File: {file_path}", fg="yellow")


def log_cli_command(command):
    click.secho(f'  ⬆️  CLI command: {" ".join(command)}', fg="bright_cyan")


def log_cli_response(result):
    click.secho(f"  ⬇️  CLI result:", fg="bright_cyan")
    click.secho(f"{result.strip()}", fg="bright_black")


def log_task(step):
    click.secho(f"📋 {step}", fg="green", bold=True)


def log_skip_task(step):
    click.secho(f"📋 {step}", fg="green", bold=True)


def log_success(step):
    click.secho(f"✅ {step}", fg="green", bold=True)


def log_info(message):
    click.secho(f"  ℹ {message}", fg="bright_black")


def log_error(message):
    click.secho(f"❌ {message}", fg="red", bold=True)



from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec, ed25519
from datetime import datetime, timedelta


def create_root_ca(common_name, key_path, cert_path):
    # Generate private key
    private_key = ec.generate_private_key(ec.SECP256R1())

    # Write private key to file
    with open(key_path, "wb") as key_file:
        key_file.write(
            private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    # Generate public certificate
    subject = issuer = x509.Name(
        [x509.NameAttribute(NameOID.COMMON_NAME, common_name)]
    )
    certificate = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.utcnow() - timedelta(days=1))
        .not_valid_after(
            # Certificate is valid for 31 years
            datetime.utcnow()
            + timedelta(days=31 * 365)
        )
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=None),
            critical=True,
        )
        .add_extension(
            x509.AuthorityKeyIdentifier.from_issuer_public_key(
                private_key.public_key()
            ),
            critical=False,
        )
        .sign(private_key, hashes.SHA256())
    )

    # Write public certificate to file
    with open(cert_path, "wb") as cert_file:
        cert_file.write(certificate.public_bytes(serialization.Encoding.PEM))


def create_intermediate_ca_csr(common_name, key_path, csr_path):
    # Generate private key for Intermediate CA
    private_key = ec.generate_private_key(ec.SECP256R1())

    # Write private key to file
    with open(key_path, "wb") as key_file:
        key_file.write(
            private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    # Generate CSR
    csr = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(
            x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, common_name)])
        )
        .sign(private_key, hashes.SHA256())
    )

    # Write CSR to file
    with open(csr_path, "wb") as csr_file:
        csr_file.write(csr.public_bytes(serialization.Encoding.PEM))


def sign_intermediate_ca_csr(
    root_key_path, root_cert_path, intermediate_csr_path, intermediate_cert_path
):
    # Load Root CA's private key
    with open(root_key_path, "rb") as key_file:
        root_private_key = serialization.load_pem_private_key(
            key_file.read(),
            password=None,
        )

    # Load Root CA's certificate
    with open(root_cert_path, "rb") as cert_file:
        root_cert = x509.load_pem_x509_certificate(cert_file.read())

    # Load Intermediate CA's CSR
    with open(intermediate_csr_path, "rb") as csr_file:
        intermediate_csr = x509.load_pem_x509_csr(csr_file.read())

    # Generate Intermediate CA's certificate
    intermediate_cert = (
        x509.CertificateBuilder()
        .subject_name(intermediate_csr.subject)
        .issuer_name(root_cert.subject)
        .public_key(intermediate_csr.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.utcnow() - timedelta(days=1))
        .not_valid_after(
            # Certificate is valid for 10 years
            datetime.utcnow()
            + timedelta(days=3650)
        )
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=0),
            critical=True,
        )
        .add_extension(
            x509.AuthorityKeyIdentifier.from_issuer_public_key(
                root_private_key.public_key()
            ),
            critical=False,
        )
        .add_extension(
            x509.SubjectKeyIdentifier.from_public_key(
                intermediate_csr.public_key()
            ),
            critical=False,
        )
        .sign(root_private_key, hashes.SHA256())
    )

    # Write Intermediate CA's certificate to file
    with open(intermediate_cert_path, "wb") as cert_file:
        cert_file.write(
            intermediate_cert.public_bytes(serialization.Encoding.PEM)
        )


def create_end_entity_csr(common_name, key_path, csr_path):
    # Generate private key for End-Entity
    private_key = ec.generate_private_key(ec.SECP256R1())

    # Write private key to file
    with open(key_path, "wb") as key_file:
        key_file.write(
            private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    # Generate CSR
    csr = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(
            x509.Name(
                [
                    x509.NameAttribute(NameOID.COMMON_NAME, common_name),
                ]
            )
        )
        .sign(private_key, hashes.SHA256())
    )

    # Write CSR to file
    with open(csr_path, "wb") as csr_file:
        csr_file.write(csr.public_bytes(serialization.Encoding.PEM))


def sign_end_entity_csr(
    intermediate_key_path,
    intermediate_cert_path,
    end_entity_csr_path,
    end_entity_cert_path,
):
    # Load Intermediate CA's private key
    with open(intermediate_key_path, "rb") as key_file:
        intermediate_private_key = serialization.load_pem_private_key(
            key_file.read(),
            password=None,
        )

    # Load Intermediate CA's certificate
    with open(intermediate_cert_path, "rb") as cert_file:
        intermediate_cert = x509.load_pem_x509_certificate(cert_file.read())

    # Load End-Entity CSR
    with open(end_entity_csr_path, "rb") as csr_file:
        end_entity_csr = x509.load_pem_x509_csr(csr_file.read())

    # Generate End-Entity certificate
    end_entity_cert = (
        x509.CertificateBuilder()
        .subject_name(end_entity_csr.subject)
        .issuer_name(intermediate_cert.subject)
        .public_key(end_entity_csr.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.utcnow() - timedelta(days=1))
        .not_valid_after(
            # Certificate is valid for 1 year
            datetime.utcnow()
            + timedelta(days=365)
        )
        .add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        )
        .add_extension(
            x509.AuthorityKeyIdentifier.from_issuer_public_key(
                intermediate_private_key.public_key()
            ),
            critical=False,
        )
        .add_extension(
            x509.SubjectKeyIdentifier.from_public_key(
                end_entity_csr.public_key()
            ),
            critical=False,
        )
        .sign(intermediate_private_key, hashes.SHA256())
    )

    # Write End-Entity certificate to file
    with open(end_entity_cert_path, "wb") as cert_file:
        cert_file.write(
            end_entity_cert.public_bytes(serialization.Encoding.PEM)
        )


def read_ca_serial_number(cert_path):
    with open(cert_path, "rb") as cert_file:
        ca_cert = x509.load_pem_x509_certificate(cert_file.read())

    serial_number = ca_cert.serial_number
    return serial_number


def create_ed25519_keys(private_key_path, public_key_path):
    private_key = ed25519.Ed25519PrivateKey.generate()
    public_key = private_key.public_key()

    private_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    public_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

    with open(private_key_path, "wb") as private_key_file:
        private_key_file.write(private_pem)

    with open(public_key_path, "wb") as public_key_file:
        public_key_file.write(public_pem)


def convert_ed25519_private_pem_to_raw(pem_path):
    # Read the PEM file
    with open(pem_path, "rb") as pem_file:
        pem_data = pem_file.read()

    # Load the private key from the PEM data
    private_key = serialization.load_pem_private_key(pem_data, password=None)

    # Ensure the key is an Ed25519 private key
    if not isinstance(private_key, ed25519.Ed25519PrivateKey):
        raise ValueError("The provided key is not an Ed25519 private key.")

    # Convert the private key to raw bytes
    raw_private_key = private_key.private_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PrivateFormat.Raw,
        encryption_algorithm=serialization.NoEncryption(),
    )

    return raw_private_key


def convert_ed25519_public_pem_to_raw(pem_path):
    # Read the PEM file
    with open(pem_path, "rb") as pem_file:
        pem_data = pem_file.read()

    # Load the public key from the PEM data
    public_key = serialization.load_pem_public_key(pem_data)

    # Ensure the key is an Ed25519 public key
    if not isinstance(public_key, ed25519.Ed25519PublicKey):
        raise ValueError("The provided key is not an Ed25519 public key.")

    # Convert the public key to raw bytes
    raw_public_key = public_key.public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw,
    )

    return raw_public_key



import base64
import json
import os
import sys
from peridio_evk.utils import *
from peridio_evk.log import *
from peridio_evk.crypto import *

def do_create_product(name):
    log_task('Creating product')
    log_info(f'Product Name: {name}')

    evk_config = read_evk_config()
    result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'products', 'create', '--name', name, '--organization-prn', evk_config['organization_prn']])
    if result.returncode != 0:
        response = json.loads(result.stderr)
        if "has already been taken" in response['data']['params']['name']:
            log_skip_task('Product already exists')
        result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'products', 'list', '--search', f'name:\'{name}\''])
        if result.returncode == 0:
            response = json.loads(result.stdout)
            product_prn = response['products'][0]['prn']
    else:
        response = json.loads(result.stdout)
        product_prn = response['product']['prn']
    log_info(f'Product PRN: {product_prn}')
    cohorts = create_product_cohorts(product_prn, name)
    return cohorts

def create_product_cohorts(product_prn, product_name):
    evk_config = read_evk_config()
    log_task(f'Creating Product Cohorts')
    cohorts = [
        ('release', 'This cohort is for devices running stable, production-ready firmware releases that are suitable for end-users or wider deployment.'),
        ('release-debug', 'Devices in this cohort run release candidate builds with debugging features enabled, allowing for more in-depth testing and issue diagnosis in a near-production environment.'),
        ('daily-release', 'This cohort is for devices running daily release builds, which are more stable than debug builds but still updated frequently for testing and validation purposes.'),
        ('daily-debug', 'This cohort is used for devices that run daily debug builds, typically used by developers for active development and testing.')
    ]

    cohort_prns = []
    for cohort, desc in cohorts:
        log_info(f'Cohort: {cohort}')

    for cohort, desc in cohorts:
        result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'cohorts', 'create', '--name', cohort, '--description', desc, '--organization-prn', evk_config['organization_prn'], '--product-prn', product_prn])
        if result.returncode != 0:
            log_skip_task('Cohort already exists')
            result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'cohorts', 'list', '--search', f'name:\'{cohort}\''])
            if result.returncode == 0:
                response = json.loads(result.stdout)
                cohort_prn = response['cohorts'][0]['prn']
        else:
            response = json.loads(result.stdout)
            cohort_prn = response['cohort']['prn']
        ca = create_product_cohort_ca(product_name, cohort, cohort_prn)
        signing_keys = create_cohort_signing_key(cohort, cohort_prn)
        cohort_prns.append({'name': cohort, 'prn': cohort_prn, 'signing_keys': signing_keys, 'ca': ca})
    return cohort_prns

def create_product_cohort_ca(product_name, cohort_name, cohort_prn):
    config_path = get_config_path()
    evk_config = read_evk_config()
    root_ca_path = os.path.join(config_path, 'evk-data', 'ca')
    root_ca_key = os.path.join(root_ca_path, 'root-private-key.pem')
    root_ca_cert = os.path.join(root_ca_path, 'root-certificate.pem')

    intermediate_ca_path = os.path.join(root_ca_path, product_name, cohort_name)
    intermediate_ca_key = os.path.join(intermediate_ca_path, f'intermediate-private-key.pem')
    intermediate_ca_csr = os.path.join(intermediate_ca_path, f'intermediate-signing-request.pem')
    intermediate_ca_cert = os.path.join(intermediate_ca_path, f'intermediate-certificate.pem')
    if not os.path.exists(intermediate_ca_cert):
        log_task(f'Creating Intermediate CA')
        os.makedirs(intermediate_ca_path)
        create_intermediate_ca_csr(f'Intermediate CA {product_name} {cohort_name}', intermediate_ca_key, intermediate_ca_csr)
        log_modify_file(intermediate_ca_key)
        log_modify_file(intermediate_ca_csr)
        sign_intermediate_ca_csr(root_ca_key, root_ca_cert, intermediate_ca_csr, intermediate_ca_cert)
        log_modify_file(intermediate_ca_cert)
    else:
        log_skip_task('Intermediate CA already exists')
        log_info(f'Intermediate CA Certificate: {intermediate_ca_cert}')
        log_info(f'Intermediate CA Private-Key: {intermediate_ca_key}')

    log_task(f'Registering Intermediate CA')
    ca_certificate_serial = str(read_ca_serial_number(intermediate_ca_cert))
    result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'ca-certificates', 'list', '--search', f'description~\'Serial: {ca_certificate_serial}\''])
    response = json.loads(result.stdout)
    ca_certificate_exists = len(response['ca_certificates']) == 1
    if not ca_certificate_exists:
        log_task(f'Generating CA Certificate Verification Code')
        result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'ca-certificates', 'create-verification-code'])
        if result.returncode == 0:
            response = json.loads(result.stdout)
            verification_code = response['verification_code']
            log_info(f'Verification Code: {verification_code}')

        log_task(f'Signing Verification Certificate')
        verification_ca_key = os.path.join(intermediate_ca_path, 'verification-private-key.pem')
        verification_ca_csr = os.path.join(intermediate_ca_path, 'verification-signing-request.pem')
        verification_ca_cert = os.path.join(intermediate_ca_path, 'verification-certificate.pem')
        create_end_entity_csr(verification_code, verification_ca_key, verification_ca_csr)
        log_modify_file(verification_ca_key)
        log_modify_file(verification_ca_csr)
        sign_end_entity_csr(intermediate_ca_key, intermediate_ca_cert, verification_ca_csr, verification_ca_cert)
        log_modify_file(verification_ca_cert)

        print(intermediate_ca_cert)

        result = peridio_cli(['peridio', '--profile', evk_config['profile'],
                              'ca-certificates', 'create', '--certificate-path',
                              intermediate_ca_cert, '--verification-certificate-path',
                              verification_ca_cert, '--description', f'Intermediate CA: {product_name}:{cohort_name}. Serial: {ca_certificate_serial}'])
                              # '--jitp-cohort-prn', cohort_prn, '--jitp-product-name', product_name, '--jitp-tags', 'JITP', '--jitp-description', 'JITP', '--jitp-target', 'arm64-v8'])
        if result.returncode != 0:
            log_error(result.stderr)
            sys.exit()
    else:
        log_skip_task(f'Intermediate CA Already Registered')

    return {'certificate': intermediate_ca_cert, 'private_key': intermediate_ca_key}

def create_cohort_signing_key(cohort_name, cohort_prn):
    config_path = get_config_path()
    evk_config = read_evk_config()
    signing_keys_path = os.path.join(config_path, 'evk-data', 'signing-keys')
    if not os.path.exists(signing_keys_path):
        os.makedirs(signing_keys_path)

    cohort_public_key_pem = os.path.join(signing_keys_path, f'{cohort_name}-public-key.pem')
    cohort_private_key_pem = os.path.join(signing_keys_path, f'{cohort_name}-private-key.pem')

    if not os.path.exists(cohort_private_key_pem):
        log_task(f'Creating Binary Signing Key')
        create_ed25519_keys(cohort_private_key_pem, cohort_public_key_pem)
        log_modify_file(cohort_private_key_pem)
        log_modify_file(cohort_public_key_pem)
    else:
        log_skip_task(f'Binary Signing Key Already Exists')
        log_info(f'Private Key PEM: {cohort_private_key_pem}')
        log_info(f'Public Key PEM: {cohort_public_key_pem}')

    public_key_raw = convert_ed25519_public_pem_to_raw(cohort_public_key_pem)
    public_key_raw_encoded = base64.b64encode(public_key_raw).decode('utf-8')
    result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'signing-keys', 'list', '--search', f'value:\'{public_key_raw_encoded}\''])
    signing_key_name = f'{cohort_name}-signing-key'

    if result.returncode == 0:
        response = json.loads(result.stdout)
        if response['signing_keys'] == []:
            log_task(f'Registering Binary Signing Key')
            result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'signing-keys', 'create', '--organization-prn', evk_config['organization_prn'], '--value', public_key_raw_encoded, '--name', signing_key_name])

            if result.returncode == 0:
                response = json.loads(result.stdout)
                signing_key_prn = response['signing_key']['prn']
            else:
                log_error(f'A signing key named \'{signing_key_name}\' already exists in Peridio CLoud, please either rename or remove the pre-existing key via Web Console, CLI, or API.')
                sys.exit()

        else:
            log_skip_task(f'Binary Signing Key Already Registered')
            signing_key_prn = response['signing_keys'][0]['prn']
    else:
        log_error(result.stderr)

    log_task(f'Adding Signing Key to CLI Keychain')
    config_file = os.path.join(config_path, 'config.json')
    config = read_json_file(config_file)
    update_config_signing_key_pairs(config, signing_key_name, signing_key_prn, cohort_private_key_pem)
    write_json_file(config_file, config)
    return {'public_key_pem': cohort_public_key_pem, 'private_key_pem': cohort_private_key_pem}


def update_config_signing_key_pairs(config, signing_key_pair_name, signing_key_prn, signing_key_private_path):
    if 'signing_key_pairs' not in config:
        config['signing_key_pairs'] = {}
    config['signing_key_pairs'][signing_key_pair_name] = {'signing_key_prn': signing_key_prn, 'signing_key_private_path': signing_key_private_path}



import click
import os
import json
import platform
import shutil
import subprocess
import queue
import threading
import stat
import sys
from datetime import datetime
from peridio_evk.log import (
    log_cli_command,
    log_cli_response,
    log_modify_file,
    log_error,
    log_info,
)


class SubprocessResult:
    def __init__(self, stdout, stderr, returncode):
        self.stdout = stdout
        self.stderr = stderr
        self.returncode = returncode


def read_json_file(file_path):
    if os.path.exists(file_path):
        with open(file_path, "r") as f:
            content = f.read().strip()
            if not content:
                return {}
            return json.loads(content)
    else:
        return {}


def write_json_file(file_path, content):
    with open(file_path, "w") as f:
        json.dump(content, f, indent=4)
        log_modify_file(file_path)


def get_config_path():
    config_dir = os.getenv("PERIDIO_CONFIG_DIRECTORY")

    if config_dir is None:
        system = platform.system()
        if system == "Linux":
            config_dir = os.path.join(
                os.path.expanduser("~"), ".config", "peridio"
            )
        elif system == "Windows":
            from ctypes import windll, create_unicode_buffer

            CSIDL_APPDATA = 0x001A
            buf = create_unicode_buffer(1024)
            windll.shell32.SHGetFolderPathW(None, CSIDL_APPDATA, None, 0, buf)
            config_dir = os.path.join(buf.value, "peridio")
        elif system == "Darwin":
            config_dir = os.path.join(
                os.path.expanduser("~"),
                "Library",
                "Application Support",
                "peridio",
            )
        else:
            raise RuntimeError(f"Unsupported platform: {system}")

    if not os.path.exists(config_dir):
        os.makedirs(config_dir)

    return config_dir


def get_evk_config_path():
    config_path = get_config_path()
    return os.path.join(config_path, "evk.json")


def read_evk_config():
    evk_config_path = get_evk_config_path()
    return read_json_file(evk_config_path)


def peridio_cli(command):
    if shutil.which("peridio") is None:
        log_error('"peridio" CLI executable not found in the system PATH.')
        raise click.ClickException(
            "Please install the peridio CLI and ensure it is available in the system PATH."
        )
    log_cli_command(command)

    try:
        # Start the subprocess
        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,  # Ensures the output is in text mode
        )

        # Queues to collect stdout and stderr output
        stdout_queue = queue.Queue()
        stderr_queue = queue.Queue()

        # Start threads to stream stdout and stderr
        stdout_thread = threading.Thread(
            target=stream_stdout, args=(process.stdout, stdout_queue)
        )
        stderr_thread = threading.Thread(
            target=stream_stderr, args=(process.stderr, stderr_queue)
        )
        stdout_thread.start()
        stderr_thread.start()

        # Wait for the process to finish
        process.wait()

        # Wait for the threads to finish
        stdout_thread.join()
        stderr_thread.join()

        # Collect stdout output from the queue
        stdout_output = ""
        while not stdout_queue.empty():
            stdout_output += stdout_queue.get()

        # Collect stderr output from the queue
        stderr_output = ""
        while not stderr_queue.empty():
            stderr_output += stderr_queue.get()

        return_code = process.returncode
        pretty_print_result = str(stdout_output)

        if return_code == 0 and pretty_print_result != "":
            formatted_result = json.loads(pretty_print_result)
            log_cli_response(json.dumps(formatted_result, indent=2))

        result = SubprocessResult(
            str(stdout_output), str(stderr_output), return_code
        )
        return result

    except Exception as e:
        log_error(e.with_traceback)
        return SubprocessResult(None, None, None)


def generate_random_bytes_file(file_path, length):
    # Generate random bytes
    random_bytes = os.urandom(length)

    # Write the bytes to a file
    with open(file_path, "wb") as file:
        file.write(random_bytes)


def stream_stdout(pipe, output_queue):
    for line in iter(pipe.readline, ""):
        if line:
            output_queue.put(line)
    pipe.close()


def stream_stderr(pipe, output_queue):
    for line in iter(pipe.readline, ""):
        if line:
            click.secho(f"     CLI: {line.strip()}", fg="bright_black")
            output_queue.put(line)
    pipe.close()


def find_dict_by_name(list_of_dicts, name_value):
    return next((d for d in list_of_dicts if d.get("name") == name_value), None)


def get_current_time_iso8601():
    # Get the current UTC time
    now = datetime.utcnow()

    # Format the time to ISO 8601 format with a 'Z' at the end to denote UTC time
    formatted_time = now.strftime("%Y-%m-%dT%H:%M:%SZ")

    return formatted_time


def boolean_to_string_lower(value):
    if isinstance(value, bool):
        return str(value).lower()
    else:
        raise ValueError("Input is not a boolean value.")


def write_file_x(file_path, content):
    with open(file_path, "w") as file:
        file.write(content)
    st = os.stat(file_path)
    os.chmod(file_path, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)


def filter_dicts(dict_list, key, value):
    return [
        dictionary for dictionary in dict_list if dictionary.get(key) == value
    ]


def get_docker_client():
    try:
        import docker

        client = docker.from_env()
        # Test the connection to Docker service
        client.ping()
        log_info("Using Docker client")
        return client
    except:
        return None


def get_container_client():
    # Check if Docker is installed
    try:
        subprocess.run(
            ["docker", "--version"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        client = get_docker_client()
        if client:
            return client
    except (subprocess.CalledProcessError, FileNotFoundError):
        pass

    log_error("Docker needs to be running to start device containers")
    sys.exit()

def sort_dict_keys(d):
    if isinstance(d, dict):
        # Recursively sort the dictionary
        return {k: sort_dict_keys(v) for k, v in sorted(d.items())}
    elif isinstance(d, list):
        # Recursively sort each element in the list
        return [sort_dict_keys(item) for item in d]
    else:
        return d



import uboot


def create_uboot_env(env_vars, env_file_path, env_size):
    """
    Create a U-Boot environment file with a valid CRC using the uboot Python module and specify the environment size.

    Parameters:
    env_vars (dict): A dictionary of environment variables.
    env_file_path (str): The path where the environment file will be saved.
    env_size (int): The desired size of the environment file in bytes.
    """
    # Create an instance of the U-Boot environment
    env = uboot.EnvBlob(size=env_size)

    # Set the environment variables
    for key, value in env_vars.items():
        env.set(key, value)

    # Export the environment data
    env_data = env.export()

    # Check if the environment data is smaller than the specified size
    if len(env_data) > env_size:
        raise ValueError("Environment data exceeds the specified size.")

    # Pad the environment data to the desired size
    padded_env_data = env_data.ljust(env_size, b"\x00")

    # Write the padded environment data to a file
    with open(env_file_path, "wb") as f:
        f.write(padded_env_data)



import click
import os
from peridio_evk.utils import *
from peridio_evk.crypto import create_root_ca
from peridio_evk.log import log_task, log_modify_file, log_info, log_skip_task
from peridio_evk.product import do_create_product
from peridio_evk.releases import do_create_artifacts
from peridio_evk.commands.devices import (
    do_create_device_environments,
    do_create_device_certificates,
    do_register_devices,
    devices,
)


@click.command()
@click.option(
    "--organization-name",
    required=True,
    type=str,
    help="Name of the organization",
)
@click.option(
    "--organization-prn",
    required=True,
    type=str,
    help="PRN of the organization",
)
@click.option(
    "--api-key", required=True, type=str, help="API key for authentication"
)
@click.option(
    "--product-name",
    required=False,
    type=str,
    default="edge-inference",
    help="Product Name (Optional)",
)
def initialize(organization_name, organization_prn, api_key, product_name):
    log_task("Initializing EVK")
    log_info(f"Organization Name: {organization_name}")
    log_info(f"Organization PRN: {organization_prn}")
    log_info(f"Product Name: {product_name}")
    log_info(f"API key: {api_key}")

    if click.confirm(
        "Running this task may take several minutes to complete.\nYou may run this task over again in the case of errors as it will not duplicate data\n\nProceed?",
        default=False,
    ):
        do_initialize(organization_name, organization_prn, api_key)
        cohort_prns = do_create_product(product_name)
        release_cohort = find_dict_by_name(cohort_prns, "release")
        release_cohort_prn = release_cohort["prn"]
        release, artifacts = do_create_artifacts(organization_prn, release_cohort_prn)
        do_create_device_environments(devices, release, artifacts, cohort_prns)
        updated_devices = do_create_device_certificates(
            devices, release_cohort["ca"]
        )
        filtered_devices = filter_dicts(updated_devices, "tags", ["canary"])
        do_register_devices(filtered_devices, product_name, release_cohort_prn)


def do_initialize(organization_name, organization_prn, api_key):
    log_task("Updating CLI and EVK configuration")
    config_path = get_config_path()
    check_default_cli_config(config_path)
    config_file = os.path.join(config_path, "config.json")
    config = read_json_file(config_file)
    update_config(config, organization_name)
    write_json_file(config_file, config)

    credentials_file = os.path.join(config_path, "credentials.json")
    credentials = read_json_file(credentials_file)
    update_credentials(credentials, organization_name, api_key)
    write_json_file(credentials_file, credentials)

    evk_config = read_evk_config()
    evk_config_file = get_evk_config_path()
    update_evk_config(evk_config, organization_name, organization_prn)
    write_json_file(evk_config_file, evk_config)

    root_ca_path = os.path.join(config_path, "evk-data", "ca")
    root_ca_key = os.path.join(root_ca_path, "root-private-key.pem")
    root_ca_cert = os.path.join(root_ca_path, "root-certificate.pem")

    if not os.path.exists(root_ca_path):
        log_task(f"Creating Root CA")
        os.makedirs(root_ca_path)
        create_root_ca(
            f"Root CA {organization_name}", root_ca_key, root_ca_cert
        )
        log_modify_file(root_ca_key)
        log_modify_file(root_ca_cert)
    else:
        log_skip_task("Root CA already exists")
        log_info(f"Root CA Certificate: {root_ca_cert}")
        log_info(f"Root CA Private-Key: {root_ca_key}")

    # Test that the 'peridio' executable is configured by calling the system
    log_task(f"Verifying CLI configuration")
    profile_name = organization_name
    peridio_cli(["peridio", "--profile", profile_name, "users", "me"])


def update_config(config, organization_name):
    if "profiles" not in config:
        config["profiles"] = {}
    config["profiles"][organization_name] = {
        "organization_name": organization_name
    }


def update_credentials(credentials, organization_name, api_key):
    credentials[organization_name] = {"api_key": api_key}


def update_evk_config(evk_config, organization_name, organization_prn):
    evk_config["profile"] = organization_name
    evk_config["organization_name"] = organization_name
    evk_config["organization_prn"] = organization_prn


def check_default_cli_config(config_path):
    default_cli_config = {"version": 1, "profiles": {}, "signing_key_pairs": {}}

    default_credentials = {}
    config_file_path = os.path.join(config_path, "config.json")
    credentials_file_path = os.path.join(config_path, "credentials.json")
    if not os.path.exists(config_path):
        log_task(f"Creating CLI Config")
        os.makedirs(config_path)
    if not os.path.exists(config_file_path):
        log_modify_file(config_file_path)
        write_json_file(config_file_path, default_cli_config)
    if not os.path.exists(credentials_file_path):
        log_modify_file(credentials_file_path)
        write_json_file(credentials_file_path, default_credentials)



from setuptools import setup, find_packages

setup(
    name="peridio_evk",
    version="0.1.7",
    author="Peridio Developers",
    author_email="support@peridio.com",
    description="The Peridio Evaluation Kit",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/peridio/peridio-evk",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    install_requires=["click", "cryptography", "uboot", "docker"],
    entry_points={
        "console_scripts": [
            "peridio-evk = peridio_evk.cli:cli",
        ],
    },
)



import click
from peridio_evk.commands.initialize import initialize
from peridio_evk.commands.devices import (
    devices_start,
    devices_stop,
    device_attach,
)


@click.group()
def cli():
    pass


cli.add_command(initialize)
cli.add_command(devices_start)
cli.add_command(devices_stop)
cli.add_command(device_attach)

if __name__ == "__main__":
    cli()



import click
import sys
import select
import termios
import tty
import uuid
import json
import hashlib
import base64
from peridio_evk.log import *
from peridio_evk.utils import *
from peridio_evk.uboot_env import *
from peridio_evk.crypto import *

devices = [
    {'identifier': 'EI-ML-0001', 'target': 'arm64-v8', 'tags': ['canary']},
    {'identifier': 'EI-ML-0002', 'target': 'arm64-v8', 'tags': ['canary']},
    {'identifier': 'EI-ML-0003', 'target': 'arm64-v8', 'tags': []},
    {'identifier': 'EI-ML-0004', 'target': 'arm64-v8', 'tags': []},
    {'identifier': 'EI-ML-0005', 'target': 'arm64-v8', 'tags': []},
    {'identifier': 'EI-ML-0006', 'target': 'arm64-v8', 'tags': []}
]

peridio_json_template = {
  "version": 1,
  "fwup": {
    "devpath": "/etc/peridiod/peridio.fwup.img",
  },
  "cache_dir": "/etc/peridiod/cache",
  "release_poll_enabled": True,
  "release_poll_interval": 5000,
  "remote_shell": True,
  "targets": ["arm64-v8", "arm-ethos-u65"],
  "remote_access_tunnels": {
    "enabled": True,
    "service_ports": [22],
    "hooks": {
      "pre_up": "/etc/peridiod/hooks/pre-up.sh",
      "pre_down": "/etc/peridiod/hooks/pre-down.sh"
    }
  },
  "node": {
    "key_pair_source": "file",
    "key_pair_config": {
      "private_key_path": "/etc/peridiod/device-private-key.pem",
      "certificate_path": "/etc/peridiod/device-certificate.pem"
    }
  }
}

peridio_rat_pre_up = """
#!/usr/bin/env bash
#
# Args
# 1: Wireguard Network Interface Name
# 2: Destination service port number

set -e

IFNAME=$1
DPORT=$2

COUNTER_FILE="/tmp/peridio_counter_${DPORT}"

if [[ ! -f "$COUNTER_FILE" ]]; then
  echo 0 > "$COUNTER_FILE"
fi

# Read the current counter value
COUNTER=$(cat "$COUNTER_FILE")

# If its the first connection, start the ssh service
if [ "$COUNTER" -le 0 ]; then
  case $DPORT in
    22)
      exec /usr/sbin/sshd
      ;;
    *)
      ;;
  esac
fi

# Increment the counter
COUNTER=$((COUNTER + 1))

# Write the updated counter back to the file
echo "$COUNTER" > "$COUNTER_FILE"
"""

peridio_rat_pre_down = """
#!/usr/bin/env bash
#
# Args
# 1: Wireguard Network Interface Name
# 2: Destination service port number

set -e

IFNAME=$1
DPORT=$2

COUNTER_FILE="/tmp/peridio_counter_${DPORT}"

if [[ ! -f "$COUNTER_FILE" ]]; then
  COUNTER=1
fi

# Read the current counter value
COUNTER=$(cat "$COUNTER_FILE")

# Decrement the counter
COUNTER=$((COUNTER + -1))

# Write the updated counter back to the file
echo "$COUNTER" > "$COUNTER_FILE"

echo "Current counter value: $COUNTER"

# If its the last connection, stop the ssh service
if [ "$COUNTER" -le 0 ]; then
  case $DPORT in
    22)
      killall sshd
      ;;
    *)
      ;;
  esac
fi
"""

custom_entrypoint = """#!/usr/bin/env bash
ssh-keygen -A
addgroup "peridio"
adduser --disabled-password --ingroup "peridio" "peridio"
echo "peridio:peridio" | chpasswd
exec "$@"
"""

fw_env = "/etc/peridiod/uboot.env 0x0000 0x20000"

@click.command(name='devices-start')
@click.option(
    "--tag",
    required=False,
    type=str,
    default="latest",
    help="peridiod image tag (Optional)",
)
def devices_start(tag):
    container_client = get_container_client()
    log_task('Starting Virtual Devices')
    image_tag = f'docker.io/peridio/peridiod:{tag}'
    log_info(f"Pulling image: {image_tag}")
    image = container_client.images.pull(image_tag)

    if not bool(image.id):
        log_error("Invalid Image Tag")
        return

    config_path = get_config_path()
    devices_path = os.path.join(config_path, 'evk-data', 'devices')

    for device in devices:
        container_name = f'peridio-{device["identifier"]}'
        try:
            container_client.containers.get(container_name)
            log_info(f'Device {device["identifier"]} container already started')
            continue
        except Exception as e:
            log_info(f'Starting Device {device["identifier"]}')
        try:
            device_path = os.path.join(devices_path, device['identifier'])
            volumes = [
                {'type': 'bind', 'source': device_path, 'target': '/etc/peridiod'},
            ]
            env_vars = {
                "PERIDIO_LOG_LEVEL": "debug"
            }
            entrypoint = ['/etc/peridiod/entrypoint.sh']
            cmd = ["/opt/peridiod/bin/peridiod", "start_iex"]
            container_client.containers.run(
                image_tag,
                stdin_open=True,
                tty=True,
                detach=True,
                mounts=volumes,
                name=container_name,
                auto_remove=True,
                environment=env_vars,
                entrypoint=entrypoint,
                cap_add=["NET_ADMIN"],
                security_opt=["disable"],
                command=cmd
            )
        except Exception as e:
          log_error(f'error {e}')

@click.command(name='devices-stop')
def devices_stop():
    container_client = get_container_client()
    log_task('Stopping Virtual Devices')
    for device in devices:
        container_name = f'peridio-{device["identifier"]}'
        try:
            container = container_client.containers.get(container_name)
            log_info(f'Stopping {device["identifier"]}')
            container.stop()
        except:
            log_info(f'Device {device["identifier"]} container already stopped')

@click.argument('device_identifier')
@click.command(name='device-attach')
def device_attach(device_identifier):
    container_client = get_container_client()
    try:
        container = container_client.containers.get(f'peridio-{device_identifier}')
        log_task(f'Attaching To Container {device_identifier}')
        exec_instance = container.exec_run('/bin/bash', tty=True, stream=True, socket=True, detach=False, stdin=True)
        old_tty_settings = termios.tcgetattr(sys.stdin)
        sock = exec_instance.output
        try:
            # Set terminal to raw mode
            tty.setraw(sys.stdin.fileno())
            log_info("Attached to the container")
            while True:
                # Wait for either input from the user or output from the container
                readable, _, _ = select.select([sys.stdin, sock], [], [])

                for r in readable:
                    if r == sys.stdin:
                        # Read user input and send it to the container's stdin
                        user_input = os.read(sys.stdin.fileno(), 1024)
                        sock._sock.send(user_input)
                    else:
                        # Read logs from the container and print them
                        output = sock._sock.recv(1024)
                        if not output:
                            return
                        output_decoded = output.decode('utf-8')
                        sys.stdout.write(output_decoded)
                        sys.stdout.flush()
        finally:
            # Restore the terminal settings
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_tty_settings)
            sock.close()

    except:
        log_info(f'Device {device_identifier} not running')



def do_create_device_environments(devices, release, artifacts, cohorts):
    config_path = get_config_path()
    devices_path = os.path.join(config_path, 'evk-data', 'devices')

    device_targets = ['arm64-v8', 'arm-ethos-u65']
    peridio_bin_installed = ''

    cohort = [cohort for cohort in cohorts if cohort.get('name') == 'release'][0]
    public_key_pem = cohort['signing_keys']['public_key_pem']
    public_key_raw = convert_ed25519_public_pem_to_raw(public_key_pem)
    public_key_raw_encoded = base64.b64encode(public_key_raw).decode('utf-8')

    for artifact in artifacts:
        sorted_custom_metadata = sort_dict_keys(artifact['custom_metadata'])
        custom_metadata = json.dumps(sorted_custom_metadata, separators=(',', ':'))
        log_info(f'Sorted: {custom_metadata}')
        custom_metadata_bytes = custom_metadata.encode('utf-8')
        hash_object = hashlib.new('sha256')
        hash_object.update(custom_metadata_bytes)
        custom_metadata_hash = hash_object.digest()
        installed_targets = [target for target in artifact['targets'] if target['target'] in device_targets]
        for target in installed_targets:
            binary_uuid = target['binary_prn'].split(':')[-1]
            log_info(f'binary id: {binary_uuid}')
            id = uuid.UUID(binary_uuid).bytes
            peridio_bin_installed = peridio_bin_installed + base64.b16encode(id).lower().decode('utf-8')
            log_info(f'id: {base64.b16encode(id).lower().decode("utf-8")}')
            peridio_bin_installed = peridio_bin_installed + base64.b16encode(custom_metadata_hash).lower().decode('utf-8')
            log_info(f'custom_metadata_hash: {base64.b16encode(custom_metadata_hash).lower().decode("utf-8")}')

    if not os.path.exists(devices_path):
        log_task(f'Creating Device Environments')
        os.makedirs(devices_path)

    for device in devices:
        device_path = os.path.join(devices_path, device['identifier'])
        if not os.path.exists(device_path):
            os.makedirs(device_path)

        device_env = {
            'peridio_rel_current': release['prn'],
            'peridio_vsn_current': release['version'],
            'peridio_bin_current': peridio_bin_installed
        }

        device_env_path = os.path.join(device_path, 'uboot.env')
        env_size_hex = int('0x20000', 16)
        create_uboot_env(device_env, device_env_path, env_size_hex)

        device_fw_env_config = os.path.join(device_path, 'fw_env.config')
        with open(device_fw_env_config, "w") as fw_env_config:
            fw_env_config.write(
                fw_env
            )

        entrypoint_file = os.path.join(device_path, 'entrypoint.sh')
        write_file_x(entrypoint_file, custom_entrypoint)

        peridio_json = peridio_json_template
        peridio_json['trusted_signing_keys'] = [public_key_raw_encoded]
        peridio_json_path = os.path.join(device_path, 'peridio.json')
        with open(peridio_json_path, 'w') as file:
            file.write(json.dumps(peridio_json, indent=2))

        rat_hooks_path = os.path.join(device_path, 'hooks')
        if not os.path.exists(rat_hooks_path):
            os.makedirs(rat_hooks_path)

        pre_up_path = os.path.join(rat_hooks_path, 'pre-up.sh')
        write_file_x(pre_up_path, peridio_rat_pre_up)

        pre_down_path = os.path.join(rat_hooks_path, 'pre-down.sh')
        write_file_x(pre_down_path, peridio_rat_pre_down)

def do_create_device_certificates(devices, signer_ca):
    config_path = get_config_path()
    devices_path = os.path.join(config_path, 'evk-data', 'devices')
    if not os.path.exists(devices_path):
        log_task(f'Creating Device Environments')
        os.makedirs(devices_path)

    for device in devices:
        device_path = os.path.join(devices_path, device['identifier'])
        device_key = os.path.join(device_path, 'device-private-key.pem')
        device_csr = os.path.join(device_path, 'device-signing-request.pem')
        device_cert = os.path.join(device_path, 'device-certificate.pem')
        if not os.path.exists(device_cert):
            create_end_entity_csr(device['identifier'], device_key, device_csr)
            log_modify_file(device_key)
            log_modify_file(device_csr)
            sign_end_entity_csr(signer_ca['private_key'], signer_ca['certificate'], device_csr, device_cert)
            log_modify_file(device_cert)
        device['certificate'] = device_cert
        device['private_key'] = device_key

    return devices

def do_register_devices(devices, product_name, cohort_prn):
    evk_config = read_evk_config()
    for device in devices:
        log_task(f'Registering Device')
        log_info(f'Device Identifier: {device["identifier"]}')
        log_info(f'Device Certificate: {device["certificate"]}')
        log_info(f'Device Private Key: {device["private_key"]}')

        result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'devices',
                              'create', '--identifier', device['identifier'],
                              '--product-name', product_name, '--cohort-prn',
                              cohort_prn, '--tags', f'{",".join(device["tags"])}', '--target', device["target"]])
        if result.returncode != 0:
            log_skip_task('Device already exists')

        result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'device-certificates', 'create', '--device-identifier', device['identifier'], '--product-name', product_name, '--certificate-path', device['certificate']])
        if result.returncode != 0:
            log_skip_task('Device certificate already exists')



from peridio_evk.utils import *
from peridio_evk.log import *

def do_create_artifacts(organization_prn, cohort_prn):
    log_task('Creating Artifacts')
    artifacts_start = [
        {'name': 'edge-inference-os', 'description': 'Edge Inference Product OS', 'version': 'v1.12.1', 'targets': [{'target': 'arm64-v8', 'bytes': 67108864}, {'target': 'x86_64', 'bytes': 69206016}],
            "custom_metadata": {"peridiod": {"installer": "fwup", "installer_opts": {"cache_enabled": False}, "reboot_required": True}}},
        {'name': 'edge-inference-service', 'description': 'Edge Inference Service', 'version': 'v1.5.3', 'targets': [{'target': 'arm64-v8', 'bytes': 10485760}, {'target': 'x86_64', 'bytes': 14680064}],
            "custom_metadata": {"peridiod": {"installer": "file", "installer_opts": {"path": "/opt/edge-inference", "name": "edge-inference-service.img", "reboot_required": False}}}},
        {'name': 'edge-inference-peripheral', 'description': 'Edge Inference Peripheral Firmware', 'version': 'v1.9.10', 'targets': [{'target': 'arm-cortex-m33', 'bytes': 2097152}],
            "custom_metadata": {"peridiod": {"installer": "file", "installer_opts": {"path": "/opt/edge-inference", "name": "edge-inference-peripheral.bin", "reboot_required": False}}}},
        {'name': 'edge-inference-model', 'description': 'Edge Inference ML Model', 'version': 'v1.4.0', 'targets': [{'target': 'arm-ethos-u65', 'bytes': 33554432}],
            "custom_metadata": {"peridiod": {"installer": "file", "installer_opts": {"path": "/opt/edge-inference", "name": "edge-inference-model.onnx", "reboot_required": False}}}}
    ]

    artifacts_end = [
        {'name': 'edge-inference-os', 'description': 'Edge Inference Product OS', 'version': 'v1.12.1', 'targets': [{'target': 'arm64-v8', 'bytes': 67108864}, {'target': 'x86_64', 'bytes': 69206016}],
            "custom_metadata": {"peridiod": {"installer": "fwup", "installer_opts": {"cache_enabled": False}, "reboot_required": True}}},
        {'name': 'edge-inference-service', 'description': 'Edge Inference Service', 'version': 'v2.0.0', 'targets': [{'target': 'arm64-v8', 'bytes': 10486260}, {'target': 'x86_64', 'bytes': 14685064}],
            "custom_metadata": {"peridiod": {"installer": "file", "installer_opts": {"path": "/opt/edge-inference", "name": "edge-inference-service.img", "reboot_required": False}}}},
        {'name': 'edge-inference-peripheral', 'description': 'Edge Inference Peripheral Firmware', 'version': 'v1.9.10', 'targets': [{'target': 'arm-cortex-m33', 'bytes': 2097152}],
            "custom_metadata": {"peridiod": {"installer": "file", "installer_opts": {"path": "/opt/edge-inference", "name": "edge-inference-peripheral.bin", "reboot_required": False}}}},
        {'name': 'edge-inference-model', 'description': 'Edge Inference ML Model', 'version': 'v2.1.0', 'targets': [{'target': 'arm-ethos-u65', 'bytes': 43554432}],
            "custom_metadata": {"peridiod": {"installer": "file", "installer_opts": {"path": "/opt/edge-inference", "name": "edge-inference-model.onnx", "reboot_required": False}}}}
    ]

    bundle_start, artifacts = do_create_artifacts_bundle(artifacts_start, 'r1001', organization_prn)
    bundle_end, _artifacts = do_create_artifacts_bundle(artifacts_end, 'r1002', organization_prn)

    release_from = do_create_release('release-r1001', organization_prn, cohort_prn, bundle_start, '1.1.0', '', False, [])
    do_create_release('release-r1002', organization_prn, cohort_prn, bundle_end, '2.0.0', '~> 1.1', True, ['canary'])
    return release_from, artifacts

def do_create_artifacts_bundle(artifacts, bundle_name, organization_prn):
    for artifact in artifacts:
        log_info(f'{artifact["name"]}: {artifact["version"]}')
    artifact_version_prns = []
    for artifact in artifacts:
        evk_config = read_evk_config()
        result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'artifacts', 'create', '--organization-prn', evk_config['organization_prn'], '--name', artifact['name'], '--description', artifact['description'], '--custom-metadata', json.dumps(artifact['custom_metadata'])])
        if result.returncode != 0:
            log_skip_task('Artifact Exists')
            result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'artifacts', 'list', '--search', f'name:\'{artifact["name"]}\''])
            response = json.loads(result.stdout)
            artifact_prn = response['artifacts'][0]['prn']
        else:
            log_task('Creating Artifact')
            response = json.loads(result.stdout)
            artifact_prn = response['artifact']['prn']
            log_info(f'Artifact PRN: {artifact_prn}')

        result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'artifact-versions', 'create', '--artifact-prn', artifact_prn, '--version', artifact['version'], '--description', artifact['version']])
        if result.returncode != 0:
            log_skip_task('Artifact Version Exists')
            result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'artifact-versions', 'list', '--search', f'artifact_prn:\'{artifact_prn}\' and description:\'{artifact["version"]}\''])
            response = json.loads(result.stdout)
            artifact_version_prn = response['artifact_versions'][0]['prn']
        else:
            log_task('Creating Artifact Version')
            response = json.loads(result.stdout)
            artifact_version_prn = response['artifact_version']['prn']
            log_info(f'Artifact Version PRN: {artifact_version_prn}')

        artifact_version_prns.append(artifact_version_prn)

        config_path = get_config_path()
        artifacts_path = os.path.join(config_path, 'evk-data', 'artifacts')
        if not os.path.exists(artifacts_path):
            os.makedirs(artifacts_path)

        for target in artifact['targets']:
            artifact_binary_path = os.path.join(artifacts_path, f'{artifact["name"]}-{artifact["version"]}-{target["target"]}')
            if not os.path.exists(artifact_binary_path):
                log_task('Creating Artifact Binary')
                generate_random_bytes_file(artifact_binary_path, target['bytes'])
                log_modify_file(artifact_binary_path)
            else:
                log_skip_task('Artifact Binary Exists')

            result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'binaries', 'create', '--artifact-version-prn', artifact_version_prn, '--target', target['target'], '--content-path', artifact_binary_path, '--signing-key-pair', 'release-signing-key'])
            if result.returncode != 0:
                response = json.loads(result.stdout)
                binary_prn = response['binaries'][0]['prn']
                log_info(f'Binary PRN: {binary_prn}')
            else:
                response = json.loads(result.stdout)
                binary_prn = response['binary']['prn']
                log_info(f'Binary PRN: {binary_prn}')
            target['binary_prn'] = binary_prn

    result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'bundles', 'create', '--artifact-version-prns', f'{",".join(artifact_version_prns)}', '--name', bundle_name, '--organization-prn', organization_prn])
    if result.returncode == 0:
        log_task('Creating Bundle')
        log_info(f'Bundle Name: {bundle_name}')
        response = json.loads(result.stdout)
        bundle_prn = response['bundle']['prn']
    else:
        log_skip_task('Bundle already Exists')
        result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'bundles', 'list', '--search', f'name:\'{bundle_name}\''])
        response = json.loads(result.stdout)
        bundle_prn = response['bundles'][0]['prn']
    return bundle_prn, artifacts

def do_create_release(release_name, organization_prn, cohort_prn, bundle_prn, version, version_requirement, disabled, phase_tags):
    log_task('Create Release')
    log_info(f'Release Name: {release_name}')
    log_info(f'Cohort PRN: {cohort_prn}')
    log_info(f'Bundle PRN: {bundle_prn}')

    evk_config = read_evk_config()
    current_time = get_current_time_iso8601()

    command = ['peridio', '--profile', evk_config['profile'], 'releases', 'create', '--organization-prn', organization_prn, '--bundle-prn', bundle_prn, '--cohort-prn', cohort_prn, '--name', release_name, '--schedule-date', current_time, '--disabled', boolean_to_string_lower(disabled), '--version', version, '--version-requirement', version_requirement]

    if not phase_tags:
        command.append('--phase-value')
        command.append('1.0')
    else:
        command.append('--phase-tags')
        command.append(f'{",".join(phase_tags)}')

    result = peridio_cli(command)
    if result.returncode == 0:
        response = json.loads(result.stdout)
        release = response['release']
    else:
        result = peridio_cli(['peridio', '--profile', evk_config['profile'], 'releases', 'list', '--search', f'version:\'{version}\''])
        if result.returncode == 0:
            response = json.loads(result.stdout)
            release = response['releases'][0]
    return release
